#!/usr/bin/perl

#
#  phd2control.pl
#
#  Copyright (c) 2016-2023 Andy Galasso
#  All rights reserved.
#
#  This source code is distributed under the following "BSD" license
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#    Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#    Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#    The name of Andy Galasso may not be used to endorse or promote products
#    derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.
#

use strict;
use Getopt::Std;
use IO::Socket;

my $VERSION = "0.9";

my $phd_host = "127.0.0.1";
my $phd_port = 4400;

my $sock;

sub usage {
    print <<'EOF';
phd2control - control PHD2

Usage:
  phd2control [OPTIONS] command arg...
  phd2control guide  settle_pixels settle_time settle_timeout    - start guiding wait for guiding to settle
  phd2control dither dither_pixels settle_pixels settle_time settle_timeout    - dither and wait for guiding to settle
  phd2control stop                                               - stop guiding or looping exposures
  phd2control connect                                            - connect equipment
  phd2control disconnect                                         - disconnect equipment
  phd2control appstate                                           - check application state
  phd2control pause [full]                                       - pause guiding
  phd2control resume                                             - resume guiding
  phd2control set_lock_position X Y                              - set the current lock position
  phd2control shutdown                                           - shutdown phd2
  phd2control listen                                             - display phd2 events
  phd2control any METHOD [PARAMS | @FILENAME]                    - call server method METHOD with parameters (JSON)
                                                                   from the command-line, or in the named file (excluding
                                                                   the @-sign)
  phd2control version                                            - display phd2control version

Options:
  -h PHD2_HOST                          - hostname or IP address of machine running PHD2 (default: localhost)
  -i PHD2_INSTANCE                      - which PHD2 instance (default: 1)
  -o OUTPUT_FILE                        - write command status to file OUTPUT_FILE. Command status is a single line
                                          starting with 'OK' or 'ERROR' followed by a status message

Examples:
  Start guiding and wait for guiding to settle to within 1.5 pixels for at least 8 seconds:

      phd2control guide 1.5 8 300

  Dither PHD2 by up to 3 pixels, then wait for guiding to settle

      phd2control dither 3 1.5 8 300

  Stop guiding

      phd2control stop
EOF

    exit 1;
}

sub connect_phd {
    $sock = IO::Socket::INET->new(Proto     => "tcp",
                                  PeerAddr  => $phd_host,
                                  PeerPort  => $phd_port,
        ) || die "Cannot connect to PHD2 at $phd_host:$phd_port. Is PHD2 running and server enabled?\n";
}

sub wait_resp {
    while (1) {
        my $line = <$sock>;
        die "PHD2 exited\n" unless defined $line;
        return $line if $line =~ /"jsonrpc".*,"result":.*,"id":1[^\d]/;
        if ($line =~ /"jsonrpc".*,"error":(.*),"id":((1[^\d])|null)/) {
            my $err = $1;
            print STDERR $err, "\n";
            $err =~ s/.*"message":"(.*)".*/\1/;
            die $err, "\n";
        }
    }
}

sub wait_settle_done {
    while (1) {
        my $line = <$sock>;
        die "PHD2 exited\n" unless defined $line;
        next if $line =~ /"Event":"GuideStep"/;
        print $line;
        next unless $line =~ /"Event":"SettleDone"/;
        if ($line =~ /"Error":"([^"]+)"/) {
            die $1, "\n";
        }
        return;
    }
}

sub do_guide {
    my ($settle_px, $settle_time, $timeout, $x, $y, $w, $h) = @_;
    usage() unless $settle_px >= 0 && $settle_time >= 0 && $timeout > 5;
    if ($x =~ m/^(\d+)\D(\d+)[^\d+-]([+-]?\d+)[^\d+-]([+-]?\d+)$/) { # WxH@X,Y
        ($w,$h,$x,$y) = ($1,$2,$3,$4);
    }
    my $roi = defined($h) && qq(,"roi":[$x,$y,$w,$h]);
    connect_phd();
    print $sock qq({"method":"guide","params":{"settle":{"pixels":$settle_px,"time":$settle_time,"timeout":$timeout}$roi},"id":1}\r\n);
    wait_resp();
    wait_settle_done();
    return "guiding started";
}

sub do_dither {
    my ($dither_px, $settle_px, $settle_time, $timeout) = @_;
    usage() unless $dither_px >= 0.1 && $settle_time >= 0 && $timeout > 5;
    connect_phd();
    print $sock qq({"method":"dither","params":[$dither_px,false,{"pixels":$settle_px,"time":$settle_time,"timeout":$timeout}],"id":1}\r\n);
    wait_resp();
    wait_settle_done();
    return "dither complete";
}

sub do_stop {
    connect_phd();
    print $sock qq({"method":"stop_capture","id":1}\r\n);
    wait_resp();
    # don't return until phd2 really is stopped
    while (1) {
        print $sock qq({"method":"get_app_state","id":1}\r\n);
        my $r = wait_resp();
        last if $r =~ m("result"\s*:\s*"Stopped");
        sleep 1;
    }
    return "stopped";
}

sub do_shutdown {
    connect_phd();
    print $sock qq({"method":"shutdown","id":1}\r\n);
    wait_resp();
    return "shutdown";
}

sub do_pause {
    my $full = shift;
    my $params;
    if ($full) {
        $params = qq(true,"full");
    }
    else {
        $params = "true";
    }
    connect_phd();
    print $sock qq({"method":"set_paused","params":[$params],"id":1}\r\n);
    wait_resp();
    return "paused";
}

sub do_resume {
    connect_phd();
    print $sock qq({"method":"set_paused","params":[false],"id":1}\r\n);
    wait_resp();
    return "resumed";
}

sub do_set_lock_position {
    my ($x, $y) = @_;
    usage() unless $x >= 0. && $y >= 0.;
    connect_phd();
    print $sock qq({"method":"set_lock_position","params":{"x":$x,"y":$y},"id":1}\r\n);
    wait_resp();
    return "set_lock_position";
}

sub do_connect {
    connect_phd();
    print $sock qq({"method":"set_connected","params":[true],"id":1}\r\n);
    wait_resp();
    return "connected";
}

sub do_any {
    my ($method, $params) = @_;
    $params = '[]' unless defined $params;
    if (substr($params, 0, 1) eq '@') {
        my $filename = substr($params, 1);
        local $/ = undef;
        open my $IN, "<", $filename
            or die "$filename: $!\n";
        $params = <$IN>;
    }
    # the entire JSONRPC command must be on a single line
    $params =~ s/\n/ /g;
    connect_phd();
    print $sock qq({"method":"$method","params":$params,"id":1}\r\n);
    my $resp = wait_resp();
    return $resp;
}

sub do_disconnect {
    connect_phd();
    print $sock qq({"method":"set_connected","params":[false],"id":1}\r\n);
    wait_resp();
    return "disconnected";
}

sub do_appstate {
    connect_phd();

    # AppState gets sent automatically upon connect
    while (1) {
        my $line = <$sock>;
        die "PHD2 exited\n" unless defined $line;
        if ($line =~ /"Event":"AppState".*,"State":"([^"]*)/) {
            my $appstate = $1;
            return $appstate;
        }
    }
}

sub do_listen {
    connect_phd();
    while (<$sock>) {
        print;
    }
}

# ==== main ====

our ($opt_h, $opt_i, $opt_o);
getopts('h:i:o:');

$| = 1;

if (defined $opt_h) {
    $phd_host = $opt_h;
}
if (defined $opt_i) {
    $phd_port = 4400 + $opt_i - 1;
}

my $cmd = lc shift @ARGV;
my $status;
my $ok = 1;

eval {

    if ($cmd eq 'guide') {
        $status = do_guide(@ARGV);
    }
    elsif ($cmd eq 'dither') {
        $status = do_dither(@ARGV);
    }
    elsif ($cmd eq 'stop') {
        $status = do_stop(@ARGV);
    }
    elsif ($cmd eq 'connect') {
        $status = do_connect(@ARGV);
    }
    elsif ($cmd eq 'disconnect') {
        $status = do_disconnect(@ARGV);
    }
    elsif ($cmd eq 'appstate') {
        $status = do_appstate(@ARGV);
    }
    elsif ($cmd eq 'pause') {
        $status = do_pause(@ARGV);
    }
    elsif ($cmd eq 'resume') {
        $status = do_resume(@ARGV);
    }
    elsif ($cmd eq 'set_lock_position') {
        $status = do_set_lock_position(@ARGV);
    }
    elsif ($cmd eq 'any') {
        $status = do_any(@ARGV);
    }
    elsif ($cmd eq 'version') {
        print "$VERSION\n";
    }
    elsif ($cmd eq 'shutdown') {
        $status = do_shutdown(@ARGV);
    }
    elsif ($cmd eq 'listen') {
        do_listen();
    }
    else {
        usage();
        $ok = undef;
    }
};
if ($@) {
    $ok = undef;
    $status = $@;
    chomp $status;
}

if (length($status)) {
    print $status, "\n";

    if (defined($opt_o)) {
        open(OUT, ">$opt_o") or die "$opt_o: $!\n";
        print OUT ($ok ? "OK" : "ERROR"), " ", $status, "\n";
        close OUT;
    }
}

exit($ok ? 0 : 1);
